/* ########################################################################

			     Smac file perl mode (basic)

   File: perl.sc
   Path: /mnt/LaCie/xcoral-3.49.61/SmacLib/perl.sc
   Description: 
   Created: sam. mai  2 15:51:30 CEST 2020
   Author: lionel fournigault
   Modified: dim. mai 10 02:21:10 CEST 2020
   Last maintained by: lionel fournigault

   RCS $Revision$ $State$
   

   ########################################################################

   Note: perl_color_region is defined in Smac file color.sc
   	 
   perl_new_line: return and indent
   perl_indent_line: Tab key to indent line
   perl_indent_region: use mark to indent rgion
	 
   To indent regexp are used:
	 
   perl_regexp_indent use keywords need to indent the next line to right
   perl_regexp_back_indent use keywords to indent the current line to left
	 

   ########################################################################

   Copyright (c) : lionel fournigault

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */

int perl_indent_step = 4;

char * perl_regexp_indent = "while\\|for\\|do\\|elsif\\|else\\|if\\|foreach\\|\
package\\|case\\|until\\|sub\\|{";
char * perl_regexp_indent2 =
  "if\\|else\\|for\\|while\\|do\\|try\\|{\\|catch";

char * perl_string = "\'.*\'\\|\".*\""; /* replace gen_string */


/* ================================================================================
   Return line number at position. First line of buffer = 0
   ================================================================================ */
int perl_no_line_at_pos ( int position ) {

  int pos = 0, n = 0;
		     
  while(pos != end_of_file()) {
    if (pos == position){
      return n;
    }
    if (the_char(pos) == '\n')
      n += 1;
    pos += 1;
  }
  return n;
}

/* ================================================================================
   Search position inside regexp. Return 1 if position is inside regexp else 0
   ================================================================================ */
int perl_pos_in_regexp ( char *regexp, int nsub, int position ) {

  int orig = current_position();
  int pos, re_begin, re_end;

  goto_char(0);
  while (1) {
    pos = re_forward_search (regexp);
    if ((pos == -1) || (pos == -2) || (pos >= end_of_file()))
      break;
    
    if (pos > position){
      goto_char (orig);
      return 0;
    }
    
    re_begin = re_match_beginning(nsub);
    re_end = re_match_end(nsub);
    
    if ( position >= re_begin && position <= re_end ) {
      /* Bingo */
      goto_char (orig);
      return 1;
    }

    if (re_end == end_of_file())
      break;
    else
      goto_char(re_end + 1);
  }
  goto_char (orig);
  return 0;
}

/* ================================================================================
   Search blinkchar backward in the buffer. Return position or -1 
   ================================================================================ */
int perl_pos_blink_char ( int blink_char, int c ) {
  
  int pos, c_char, pos_blink = -1, internal = 0;
  
  pos = current_position();
  
  if (pos < 2)
    return;
  else
    pos -= 2;
  
  if (pos == 0 && the_char(pos) == blink_char)
    pos_blink = 0;
  
  while ( pos >= 0 ){
    c_char = the_char(pos);
    
    if ( c_char == blink_char ) {
      if ( perl_pos_in_regexp ( re_perl_comment, 0, pos ) == 0 
	  && perl_pos_in_regexp ( perl_string, 0, pos ) == 0 ) {
	if (internal == 0) {
	  pos_blink = pos;
	  break;
	}
	else {
	  internal -= 1;
	}
      }
    }
    else if (c_char == c ) {
      if ( perl_pos_in_regexp ( re_perl_comment, 0, pos ) == 0 
	  && perl_pos_in_regexp ( perl_string, 0, pos ) == 0 ) {
	internal += 1;
      }
    }
    else
      ;
    pos -= 1;
  }
  return pos_blink;
}

/* ================================================================================
   Align two lines. Use after blink;
   ================================================================================ */
void perl_align_lines ( int ref_pos, int orig ) {

  int pos, c;
  int count = 0;

  if (perl_no_line_at_pos(ref_pos) == perl_no_line_at_pos(orig))
    return;

  /* line to align */
  goto_beginning_of_line();
  pos = current_position();
  while (pos != end_of_line()){
    if ( the_char(pos) == ' ' || the_char(pos) == '\t' ) {
      delete_char();
      count -= 1;
      pos = current_position();
    }
    else
      break;
  }

  /* Reference line */
  pos = ref_pos;
  if (the_char(pos) == '\n')
    pos -= 1;
  
  while (1) {
    if (the_char(pos) == '\n'){
      pos += 1;
      break;
    }
    if (pos == 0)
      break;
    pos -= 1;
  }
  
  while (1) {
    c = the_char(pos);
    if ( c == ' ' || c == '\t' ) {
      insert_char(c);
      count += 1;
    }
    else
      break;
    pos += 1;
  }
  goto_char ( orig + count );
}

/* ================================================================================
   Return first character in line (not blank space or tab)
   ================================================================================ */
char perl_first_char_in_line ( int line ) {
  
  int pos = current_position();
  int tmp_pos;
  
  if (current_line() != line)
    goto_line(line);
  
  if ( beginning_of_line() ==  end_of_line() ){
    goto_char (pos);
    return '\n';
  }
  else
    goto_beginning_of_line();

  while (1){
    if (current_position() == end_of_line()){
     goto_char (pos);
     return '\n';
    }
    if (the_char(current_position()) != ' ' 
	&& the_char(current_position()) != '\t'){
      tmp_pos = current_position();
      goto_char (pos);
      return the_char(tmp_pos);
    }
    else
      goto_next_char();
  }
}

/* ================================================================================
   Return last character in current line (not blank space or tab)
   Ignore single line comment
   ================================================================================ */
char perl_last_char_in_line() {
  
  int pos = end_of_line();
  
  if (pos == beginning_of_line()) /* empty line */
      return '\n';
  else
    pos -= 1;
      
  while (1) {
    if ( pos == beginning_of_line() )
      break;
    if (the_char(pos) == ' ' || the_char(pos)== '\t'
	|| perl_pos_in_regexp ( re_perl_comment, 0, pos ) == 1 )
      pos -= 1;
    else
      break;
  }
  return the_char(pos);
}

/* ================================================================================
   Return distance from end of line. 
   ================================================================================ */
int perl_distance_from_end_of_line (int pos) {
  
  int distance = 0, tmp_pos;
  
  if (end_of_line() == beginning_of_line())
    distance = 0;
    
  tmp_pos = end_of_line();
  while (tmp_pos != beginning_of_line()){
    if (pos == tmp_pos)
      break;
    else
      distance += 1;
    tmp_pos -= 1;
  }
  /* a partir d'ici tmp_pos = pos ou debut de ligne */
  tmp_pos = beginning_of_line();
  while (tmp_pos != end_of_line()){
    if (the_char(tmp_pos) != ' ' && the_char(tmp_pos) != '\t'){
      break;
    }
    else {
      tmp_pos += 1;
     }
  }
  if (tmp_pos > pos)
    distance -= tmp_pos - pos;

  return distance;
}

/* ================================================================================
   Align to previous line.
   ================================================================================ */
void perl_align_to_previous_line() {
  
  int pos, c;

  if (current_line() == 0)
    return;
  /* Remove ' ' or tab at the beginning of line */
  goto_beginning_of_line();
  while(current_char()==' ' || current_char()=='\t'){
    if (current_position() == end_of_line())
      break;
    delete_char();
  }
  
  /* Use indent of previous line */  
  goto_previous_line();
  goto_beginning_of_line();
  pos = current_position();
  goto_next_line();
  goto_beginning_of_line();
  
  while (strchr(" \t", c = the_char(pos))) {
    if (pos == end_of_line())
      break;
    insert_char(c);
    pos += 1;
  } 
}

/* ================================================================================
   Return pos blink line. Multi-lines parantheses before blink
   ================================================================================ */
int perl_pos_line_blink (int orig) {

  int tmp_pos,  pos = orig;
  int pos_blink = pos;

  goto_char(orig);
  /* on est sur le blink caractre { */
  /* recherche d'une parenthse fermante dans la ligne courante */

  while (1){
    goto_previous_char();
    tmp_pos = current_position();
    if (tmp_pos == beginning_of_line())
      break;
    if ( the_char(tmp_pos) == ' ' || the_char(tmp_pos) == '\t' )
      continue;
    else
      break;
  }

  if (the_char(tmp_pos) == ')') {
    goto_char(tmp_pos+1);   
    pos_blink = perl_pos_blink_char ('(', ')');
    if (pos_blink == -1)
      pos_blink = pos;
  }
  goto_char(orig);
  return pos_blink;
}

/* ================================================================================
   Align close bracket
   ================================================================================ */
void perl_align_close_bracket(int orig) {
  
  int pos, tmp_pos, pos_blink;
  
  goto_beginning_of_line();
  tmp_pos = current_position();
  while (1){
    if (tmp_pos == end_of_line())
      break;
    if (the_char(tmp_pos) == '}')
      break;
    tmp_pos += 1;
  }
  if (the_char(tmp_pos) != '}')
    return;
  
  if ( perl_pos_in_regexp( re_perl_comment, 0, tmp_pos ) == 0 ) { 
    pos_blink = perl_pos_blink_char ('{', '}');

    if (pos_blink != -1){
      pos = perl_pos_line_blink(pos_blink);
      if (pos != pos_blink)
	pos_blink = pos;
      goto_char(orig);
      perl_align_lines (pos_blink, orig);
    }
  }
}

/* ================================================================================
   Indent more if need
   ================================================================================ */
int perl_indent_more_if_need () {
  
  int first_char_position, n;
  int more_indent = 0;
  
  /* Move to previous line at the first character not blank space or tab */
  goto_previous_line();
  goto_beginning_of_line();
  
  while(current_char()==' ' || current_char()=='\t'){
    if (current_position() == end_of_line())
      break;
    goto_next_char();
  }
  
  first_char_position = current_position();
  /* indent more if need */
  if (((re_forward_search(perl_regexp_indent) == first_char_position)
       && perl_pos_in_regexp (re_perl_comment, 0, first_char_position) == 0
       && (perl_last_char_in_line() != ';') 
       && (perl_last_char_in_line() != '}'))
      || (perl_last_char_in_line() == '{')) {
    more_indent = 1;
    n = perl_indent_step;
    goto_next_line();
    goto_beginning_of_line();
    if ( perl_first_char_in_line(current_line()) != '{') {
      while (n--){
	insert_char(' ');
      }
    }
    if ( perl_first_char_in_line(current_line()) == '{' 
	&& perl_last_char_in_line() == '}' ) {
      n = perl_indent_step;
      while (n--){
	insert_char(' ');
      }
    }
  }
  else
    goto_next_line(); /* return to current line */
  return more_indent;
}

/* ================================================================================
   Indent less if need. (instruction sur 2 lignes prcdent la ligne courante)
   ================================================================================ */
void perl_indent_less_if_need () {
  
  int first_char_position;

  goto_previous_line();
  goto_beginning_of_line();
  if (perl_last_char_in_line() == ';' || perl_last_char_in_line() == '}'){
    goto_previous_line();
    goto_beginning_of_line();
    while(current_char()==' ' || current_char()=='\t'){
      if (current_position() == end_of_line())
	break;
      goto_next_char();
    }
    first_char_position = current_position();
    if (re_forward_search(perl_regexp_indent2) == first_char_position
	&& the_char(first_char_position) != '{'
	&& perl_last_char_in_line() != '{') {
      goto_next_line();
      if (perl_first_char_in_line(current_line()) != '}') {
	goto_next_line();
	perl_align_lines( first_char_position, current_position());
      }
    }
  }
}

/* ================================================================================
   Indent switch case. (basic)
   ================================================================================ */
void perl_indent_switch_case () {
  
  int pos, orig, line, first, n;

  line = current_line();
  orig = current_position();
  
  goto_beginning_of_line();
  pos = orig;
  while (1){
    if (pos == end_of_line())
      break;
    if (the_char(pos) == ' ' || the_char(pos) == '\t' )
      pos += 1;
    else {
      first = pos;
      break;
    }
  }

  pos = re_forward_search ("case\\|default");
  if ((pos == -1) || (pos == -2) || (first != pos) || (pos >= end_of_file())){
    goto_char(orig);
    return;
  }

  goto_previous_line();
  goto_beginning_of_line();
  pos = current_position();
  while (1){
    if (pos == end_of_line())
      break;
    if (the_char(pos) == ' ' || the_char(pos) == '\t' )
      pos += 1;
    else {
      first = pos;
      break;
    }
  }
  pos = re_forward_search ("break");
  if ((pos == -1) || (pos == -2) || (first != pos) || (pos >= end_of_file())) {
    goto_char(orig);
    return;
  }
  /* Ok */
  goto_next_line();
  goto_beginning_of_line();
  
  n = perl_indent_step;
  while (n--){
    if (current_char() == ' ')
      delete_char();
    else
      break;
  }
  goto_line(line);
}

/* ================================================================================
   Indent line. Entry point of main indent process.
   ================================================================================ */
void perl_indent_line ()
{

  int orig = current_position();
  int pos = orig, n, indent_m = 0, c_line;
  int c, first_char_position, distance = 0, tmp_pos, pos_blink;

  if (current_line()==0)
    return;
  
  distance = perl_distance_from_end_of_line(pos);
  
  /* check for close bracket */
  if ( perl_first_char_in_line (current_line()) == '}' 
      /*|| perl_last_char_in_line() == '}'*/ ){
    perl_align_close_bracket(orig);
    goto_char(orig);
    /* distance = perl_distance_from_end_of_line(current_position()); */
  }
  else {
    /* first align to previous line */
    perl_align_to_previous_line();
    indent_m = perl_indent_more_if_need ();
    
    c_line = current_line();
    /* Le cas d'une instruction sur 2 lignes */
    if (indent_m == 0) {
       perl_indent_less_if_need();
    }
    goto_line (c_line);
    perl_indent_switch_case ();
    /* goto_char(orig); */
    /* distance = perl_distance_from_end_of_line(current_position()); */
  }

  /* Return to initial position */
  if (distance == 0)
    goto_end_of_line();
  else {
    goto_end_of_line();
    while(distance > 0){
      if (current_position() == beginning_of_line())
	break;
      goto_previous_char();
      distance -= 1;
    }
  }
  if (current_buffer_is_colored())
    perl_color_region (beginning_of_line(), end_of_line());
}

/* ================================================================================
   Return and indent.
   ================================================================================ */
void perl_new_line ()
{
  insert_char('\n');
  perl_indent_line();
}

/* ================================================================================
   Indent region
   ================================================================================ */
void perl_indent_region() {

  int init_line = current_line();
  int init_position = current_position();
  int mark_line, i, start_line, end_line;

  if (mark_position() == -1)
    return;
  
  watch_on();
  goto_mark();
  mark_line = current_line();
  if (mark_line > init_line){
    start_line = init_line;
    end_line = mark_line;
  }
  else {
    start_line = mark_line;
    end_line = init_line;
  }
  goto_line(start_line);

  for (i=start_line; i < end_line; i++ ){
    perl_indent_line();
    goto_next_line();
  }
  perl_indent_line();
  goto_line (init_line);
  goto_beginning_of_line();

  watch_off();
  reset_mark();
}

/* ================================================================================
   Indent buffer
   ================================================================================ */
void perl_indent_buffer() {
  
  int line = current_line();
  
  set_mark(0);
  goto_end_of_file();
  
  perl_indent_region();
  
  goto_line(line);
  goto_beginning_of_line();
  color_buffer();
  
  reset_mark();
}

/* ================================================================================
   Insert the char " ' and colorize
   ================================================================================ */
void perl_insert_and_colorize ()
{
  insert_char(last_key());
  
  if (current_buffer_is_colored())
    perl_color_region (beginning_of_line(), end_of_line());
}

/* ================================================================================
   Insert, blink matched char and indent if need
   ================================================================================ */
void perl_insert_and_blink() {
  
  int orig, pos, tmp_pos;
  int c, c_char, blink_char, internal = 0;
  
  c = last_key();

  if (c == '}')
    blink_char = '{';
  else if (c == ']')
    blink_char = '[';
  else if (c == ')')
    blink_char = '(';
  else
    return;
   
  insert_char (c);
  orig = current_position();
  pos = orig;
  
  /* Check inside comment or string */ 
  if ( perl_pos_in_regexp ( re_perl_comment, 0, pos ) == 1 
      || perl_pos_in_regexp ( perl_string, 0, pos ) == 1 ) {
    return;
  }

  pos = perl_pos_blink_char (blink_char, c);

  if ( pos == -1)
    return;
  else
    blink(pos);

  if (blink_char == '{'){
    /* indent if need */
    tmp_pos = perl_pos_line_blink (pos);
    if (tmp_pos != pos)
      pos = tmp_pos;
    goto_char(orig);
    perl_align_lines (pos, orig);
  }
}

/* ================================================================================
   Perl key_def
   ================================================================================ */
{
  key_def("Perl", "\t", "perl_indent_line");
  key_def("Perl", "\r", "perl_new_line");
  key_def("Perl", " ", "perl_insert_and_colorize");
  
  key_def("Perl", "\"", "perl_insert_and_colorize");
  key_def("Perl", "\'", "perl_insert_and_colorize");

  key_def("Perl", "]", "perl_insert_and_blink");
  key_def("Perl", ")", "perl_insert_and_blink");
  key_def("Perl", "}", "perl_insert_and_blink");

  key_def("Perl", "^[i", "perl_indent_region");
  
  key_def("Perl", "^[u", "upcase_word");
  key_def("Perl", "^[l", "downcase_word");
  key_def("Perl", "^[c", "capitalize_word");
}
